@@ -1,4 +1,4 @@
-from .hylightpy import HILevelPopulations
+from .hylightpy import HIAtom
 from importlib.metadata import version
 
 __version__ = version("hylightpy")
