@@ -10,7 +10,7 @@ import os
 import importlib
 import scipy.special as sc
 
-class HILevelPopulations:
+class HIAtom:
     '''
     Compute level population for HI using the cascade matrix formalism.
     See Osterbrock & Ferland 2006, section 4.2
@@ -18,6 +18,15 @@ class HILevelPopulations:
     def __init__(self, nmax=60, recom=True, coll=True,
                  cache_path = './cache/',
                  caseB = True, verbose=False):
+        """Initialize the hydrogen  model. 
+        
+        :param nmax: Number of levels included in the modeln. Default is ``60``.
+        :param recom: Whether to include radiative recombination. Default is ``True``.
+        :param coll: Whether to include collisional excitation from the ground state. Default is ``True``.
+        :param cache_path: Path to store the cache files. Default is the current working directory ``'./cache/'``.
+        :param caseB: Whether to execute in Case B. Default is ``True``.
+        :param verbose: Whether to include diagnostic information. Default is ``False``.
+        """
         print(" ░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░      ░▒▓█▓▒░░▒▓██████▓▒░░▒▓█▓▒░░▒▓█▓▒░▒▓████████▓▒░ ", flush=True)
         print(" ░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░      ░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░  ░▒▓█▓▒░     ", flush=True)
         print(" ░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░      ░▒▓█▓▒░▒▓█▓▒░      ░▒▓█▓▒░░▒▓█▓▒░  ░▒▓█▓▒░     ", flush=True)
@@ -26,7 +35,7 @@ class HILevelPopulations:
         print(" ░▒▓█▓▒░░▒▓█▓▒░  ░▒▓█▓▒░   ░▒▓█▓▒░      ░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░  ░▒▓█▓▒░     ", flush=True)
         print(" ░▒▓█▓▒░░▒▓█▓▒░  ░▒▓█▓▒░   ░▒▓████████▓▒░▒▓█▓▒░░▒▓██████▓▒░░▒▓█▓▒░░▒▓█▓▒░  ░▒▓█▓▒░     ", flush=True)
         
-        # set maximum number of principle quantum number to be used - max allowed is 40
+        # set maximum number of principle quantum number to be used - max allowed is 150
         self.nmax = nmax
         assert self.nmax >= 5, "At least five levels are required. "
         self.nmaxcoll = nmax
@@ -49,62 +58,84 @@ class HILevelPopulations:
         else:
             print(f"Folder '{self.cache_path}' already exists.")
         
-        # Read Einstein coefficients
+        # Read Einstein A coefficients
         self.TabulatedEinsteinAs = importlib.resources.files('hylightpy.data').joinpath('Einstein_As_150.txt')  # name of the file
-        self.A                   = self.ReadTabulatedEinsteinCoefficients(self.TabulatedEinsteinAs)
+        self.A                   = self.read_tabulated_Einstein_coefficients(self.TabulatedEinsteinAs)
 
     
         # Read level-resolved recombination rates
         self.TabulatedRecombinationRates = importlib.resources.files('hylightpy.data').joinpath('h_iso_recomb_HI_150.dat')   # name of the file
-        self.Recom_table = self.ReadRecombinationRates(self.TabulatedRecombinationRates) # tabulated rates
-        self.Alpha_nl = self.FitRecombinationRates()                                     # fitting function to tabulated rates
+        self.Recom_table = self.read_recombination_rates(self.TabulatedRecombinationRates) # tabulated rates
+        self.Alpha_nl = self.fit_recombination_rates()                                     # fitting function to tabulated rates
         if verbose:
             print("Recombination rates read and fitted")
             
         # Read level-resolved collsional exicitation rates
         #self.TabulatedCollisonalExRates = TabulatedCollisionalExRates
         #self.CollEx_table = self.ReadCollisionalExRates(self.TabulatedCollisonalExRates)
-        self.q_nl = self.FitCollisionalExRates()
+        self.q_nl = self.fit_collisional_ex_rates()
         if verbose:
-            print("Collsional Excitaion Rates read and fitted")
+            print("Collisional Excitaion Rates read and fitted")
 
         # Compute cascade matrix
-        self.C    = self.ComputeCascadeMatrix()
+        self.C    = self.compute_cascade_matrix()
         if verbose:
             print("Cascade matrix class initialized ")
 
-        self.allconfigs = self.GetAllConfigs()
+        self.allconfigs = self.get_all_configs()
  
 
-    def GetAllConfigs(self):
+    def get_all_configs(self):
+        """Get all the configurations of all the atomic states (n, l).
+        
+        :return: All the configurations of the atomic states. 
+        """
         configs = []
         for nu in np.arange(1, self.nmax+1): # list of indices, 
             for lu in np.arange(nu):
-                conf_i = self.Config(n=nu, l=lu)
+                conf_i = self.config(n=nu, l=lu)
                 configs.append(conf_i)
         return configs
         
     ##################################################################    
     #                Recombination rate method                       #
     ##################################################################  
-    def AlphaA(self, LogT=4.0):
-        ''' Fit to case-A recombination coefficient at log temperature LogT'''
+    def alpha_A(self, LogT=4.0):
+        """Fit to Case A recombination coefficient at log temperature.
+        
+        :param LogT: Log10 of temperature. Default is ``4.0``.
+        :type LogT: float
+        :return: Fitted recombination rate coefficient. 
+        :rtype: float
+        """
         T      = 10.**LogT
         lamb   = 315614 / T
         alphaA = 1.269e-13 * lamb**(1.503)*(1.0+(lamb/0.522)**(0.470))**(-1.923)
         return alphaA
 
-    def AlphaB(self, LogT=4.0):
-        ''' Fit to case-B recombination coefficient at log temperature T'''        
+    def alpha_B(self, LogT=4.0):
+        """
+        Fit to Case B recombination coefficient at log temperature.
+ 
+        :param LogT: Log10 of temperature. Default is ``4.0``.
+        :type LogT: float
+        :return: Fitted recombination rate coefficient.
+        :rtype: float
+        """
         T      = 10.**LogT
         lamb   = 315614 / T
         alphaB = 2.753e-14 * lamb**(1.5)*(1.0+(lamb/2.740)**(0.470))**(-2.2324)
         return alphaB
         
-    def ReadRecombinationRates(self, fname):
-        '''
-        Read level-resolved recombination rates from ascii file fname, and return them
-        '''
+    def read_recombination_rates(self, fname):
+        """
+        Read level-resolved recombination rates from ascii file in the data folder (h_iso_recomb_HI_150.dat).
+        
+        :param fname: Name of the tabulated recombination coefficient file. 
+        :type fname: str
+        :returns: Dictionary of the recombination coefficients.
+        :rtype: dict
+        """
         
         # contents of the cloudy data file containing l-resolved recombination rates
         # the first line is a comment
@@ -135,10 +166,15 @@ class HILevelPopulations:
         LogTs      = np.linspace(0, 10, 41, endpoint=True)
         return {'LogTs':LogTs, 'recom_data':recom_data}
     
-    def ReadEffectiveCollisonalStrength(self, verbose=False):
-        '''
-        Read level-resolved collisional excitation rates from ascii file fname, and return them
-        '''
+    def read_effective_collisional_strength(self, verbose=False):
+        """
+        Read level-resolved collisional strength up to n = 5 level from ascii file in the data folder (h_coll_str.dat). 
+        
+        :param verbose: Output diagnostic information.
+        :type verbose: boolean
+        :returns: Collisional stength up to n = 5 tabulated by Anderson et al. 2000. 
+        :rtype: dict
+        """
         fpath = importlib.resources.files('hylightpy.data').joinpath('h_coll_str.dat')
 
         # columns in Anderson et al. 2000, 2002
@@ -168,14 +204,15 @@ class HILevelPopulations:
         return {'Ts':self.temps_Anderson, 'ups_data':coll_csv}
         
 
-    def FitRecombinationRates(self):
-        ''' 
-        Provide fitting function for recombination rate as a function of Log T
-        
-        The fitting funcition is of the form 
-        
+    def fit_recombination_rates(self):
+        """
+        Provide fitting function for recombination rate as a function of log10 T.
+        The fitting funcition is of the form:
         Recombination_rate(n, l, T) = 10**Alpha_nk(10**LogT)
-        '''
+        
+        :returns: Fitting function. 
+        """
+        
         def get_l_level_index(n=1, l=0):
             assert type(n) == np.int64 and type(l) == np.int64, 'n and l must be intergers.'
             assert n >= 1, 'Principle quantum number can not be smaller than 1.'
@@ -185,12 +222,12 @@ class HILevelPopulations:
             # offset by 1, since first line is a comment line
             return int((n-1)*n/2) + l # + 1
             
-        def FitRate(recom_data, LogTs, n=1, l=0):
+        def fit_rate(recom_data, LogTs, n=1, l=0):
             index   = get_l_level_index(n=n, l=l)
             rate    = interpolate.interp1d(LogTs, recom_data.iloc[index, 2:43].values,fill_value="extrapolate", bounds_error=False)   
             return rate
 
-        rates      = self.ReadRecombinationRates(self.TabulatedRecombinationRates)
+        rates      = self.read_recombination_rates(self.TabulatedRecombinationRates)
         LogTs      = rates['LogTs']
         recom_data = rates['recom_data']
         
@@ -198,20 +235,43 @@ class HILevelPopulations:
         Alpha_nl = {}
         for n in np.arange(1, nmax+1):
             for l in np.arange(n):
-                conf_i             = self.Config(n=n, l=l)
-                Alpha_nl[conf_i]   = FitRate(recom_data, LogTs, n=n, l=l) # try 10**()
+                conf_i             = self.config(n=n, l=l)
+                Alpha_nl[conf_i]   = fit_rate(recom_data, LogTs, n=n, l=l) # try 10**()
         return Alpha_nl
 
     def collisional_excitation_rate_Lebedev_Beigman(self, nu=6, nl=1, Te=1e4):
+        """
+        Collisional de-excitation rate calculated based on Lebedev & Beigman 1998 for Rydberg atoms (n >= 5). 
+        
+        :param nu: Upper level. Default is ``6``.
+        :type nu: int
+        :param nl: Lower level. Default is ``1``.
+        :type nl: int
+        :param Te: Temperature. Default is ``10000.0``.
+        :type Te: float
+        :returns: Collisional de-excitation rate cefficient q_ul.
+        :rtype: float
+        """
         Te = Te * unyt.K
         gnu = 2 * nu**2
         gnl = 2 * nl**2
         deltaE = self.Eion * (1 / nl**2 - 1 / nu**2)
-        #print(deltaE, Te)
         qul = gnu / gnl * np.exp(- deltaE / (unyt.boltzmann_constant_cgs * Te)) * self.collisional_deexcitation_rate_Lebedev_Beigman(nu=nu, nl=nl, Te=Te.value)
         return qul
 
     def collisional_deexcitation_rate_Lebedev_Beigman(self, nu=6, nl=1, Te=1e4):
+        """
+        Collisional excitation rate calculated based on Lebedev & Beigman 1998 for Rydberg atoms (n >= 5).
+
+        :param nu: Upper level. Default is ``6``.
+        :type nu: int
+        :param nl: Lower level. Default is ``1``.
+        :type nl: int
+        :param Te: Temperature. Default is ``10000.0``.
+        :type Te: float
+        :returns: Collisional excitation rate cefficient q_lu.
+        :rtype: float
+        """
         Te = Te * unyt.K
         gnu = 2 * nu**2 # 2n**2
         gnl = 2 * nl**2
@@ -222,9 +282,7 @@ class HILevelPopulations:
         return qlu
         
     def get_theta(self, Te, Z):
-        #print('temp is', Te)
         theta = unyt.boltzmann_constant_cgs * Te / (Z**2 * self.Eion)
-        #print('theta is', theta)
         return theta
 
     def En_potential(self, n):
@@ -252,16 +310,16 @@ class HILevelPopulations:
         fval = fval1 / fval2
         return fval
         
-    def FitCollisionalExRates(self):
-        ''' 
-        Provide fitting function for collisional excitation rate of the ground state. 
+    def fit_collisional_ex_rates(self):
+        """
+        Provide fitting function for collisional excitation rate from the ground state.
+        The fitting funcition is of the form:
+        Collisional_rate(n, l, T) = 10**q_lu(10**LogT)
         
-        The fitting funcition is of the form 
+        :returns: Fitting function of the collisioanl excitation rate coefficient.
+        """
         
-        Recombination_rate(n, l, T) = 10**Alpha_nk(10**LogT)
-        '''
-        
-        def FitCollExRate(ups_data, Ts, n=1, l=0):
+        def fit_coll_ex_rate(ups_data, Ts, n=1, l=0):
 
             if n <= 5:
                 #index   = get_l_level_index(n=n, l=l)
@@ -299,17 +357,15 @@ class HILevelPopulations:
                     
             return qfit
 
-        
-        ups     = self.ReadEffectiveCollisonalStrength(verbose=True)
+        ups     = self.read_effective_collisional_strength(verbose=True)
         Ts      = ups['Ts']
         upsdata = ups['ups_data']
         nmax     = self.nmaxcoll
         q_nl = {}
         for n in np.arange(2, nmax+1): # mininum level is 2
             for l in np.arange(n):
-                conf_i             = self.Config(n=n, l=l)
-                
-                coeffs             = FitCollExRate(upsdata, Ts, n=n, l=l)
+                conf_i             = self.config(n=n, l=l)
+                coeffs             = fit_coll_ex_rate(upsdata, Ts, n=n, l=l)
                 q_nl[conf_i]       = np.poly1d(coeffs) # add 10**
                 
         return q_nl
@@ -319,74 +375,56 @@ class HILevelPopulations:
     ##################################################################    
     #                Cascade matrix methods                          #
     ##################################################################
-    def TestAllLevelPops(self, nH = 1.0, ne = 1.0, LogT = 4.0, N={}):
-        '''
-        Verify whether these pop levels satisfy the equilibrium relation - Eq. 4.1
-        '''
-        #
-        nmax     = self.nmax
-        Config   = self.Config
-        Alpha_nl = self.Alpha_nl
-        A        = self.A
-        #
-        TestConfig = []
-        TestDiff   = []
-        for n in np.arange(1, nmax+1):
-            for l in np.arange(n):
-                lhs    = 0.0
-                conf   = Config(n=n, l=l)
-                lhs   += nH * ne * 10**Alpha_nl[conf](LogT)
-                #
-                for nu in np.arange(n+1, nmax+1):
-                    for lu in [l-1, l+1]:
-                        if (lu>= 0) & (lu < nu):
-                            conf_i = Config(n=nu, l=lu)
-                            lhs += N[conf_i] * A[conf_i][conf]
-                #
-                rhs = 0.0
-                for nd in np.arange(1, n):
-                    for ld in [l-1, l+1]:
-                        if (ld >= 0) & (ld < nd):
-                            conf_k = Config(n=nd, l=ld)
-                            rhs    += A[conf][conf_k]
-                    if (nd == 1) & (n==2) & (l == 0):
-                        ld      = 0
-                        conf_k  = Config(n=nd, l=ld)
-                        rhs    += A[conf][conf_k]
-                #
-                Nnl  = 0.0
-                diff = 1e2
-                if rhs > 0:
-                    Nnl  = lhs / rhs
-                    diff = (Nnl-N[conf])/N[conf] * 100.
-#                     if n < 10:
-#                         print("Conf = {0:s}, % diff = {1:1.4f}, N = {2:1.3e}".format(conf, diff, (N[conf])))
-                TestConfig.append(conf)
-                TestDiff.append(diff)
-        return {'Conf':TestConfig, 'Diff':TestDiff}
-
-    def ComputeLevelPop(self, nHII = 1.0, ne = 1.0, nHI=1.0, LogT=4.0, n=2, l=0, verbose=False):
-        '''
-        Compute level population for a given level  - implementents Eq. 4.10
-        Input: 
-           nH   = proton number density nH [cm^[-3]]
-           ne   = electron number density [[cm^-3]]
-           logT = logarithm of temperature
-           n    = principle quantum number of desired level
-           l    = angular momentum state of this level
-        '''
-
-        #
+    def compute_level_pop(self, nHII = [1e2] * unyt.unyt_array(1, 'cm**(-3)'), 
+                                ne = [1e2] * unyt.unyt_array(1, 'cm**(-3)'), 
+                                nHI = [1e-5] * unyt.unyt_array(1, 'cm**(-3)'), 
+                                temp = [1e4] * unyt.unyt_array(1, 'K'), 
+                                n=3, l=0, verbose=False):
+        """
+        Compute level population for a given level at a given density and temperature.
+        
+        :param nHII: Proton number density [cm^{-3}]. Default is [1e2] * unyt.unyt_array(1, 'cm**(-3)').
+        :type nHII: float
+        :param ne: Electron number density [cm^{-3}]. Default is [1e2] * unyt.unyt_array(1, 'cm**(-3)').
+        :type ne: float
+        :param nHI: Neutral hydrogen density [cm^{03}]. Default is [1e-5] * unyt.unyt_array(1, 'cm**(-3)'). If collsional excitation is not enabled, any number input here will not go into the calculation. 
+        :type nHI: float
+        :param temp: Temperature. Default is [1e4] * unyt.unyt_array(1, 'K').
+        :type temp: float
+        :param n: Principle quantum number of the desired level. Default is ``2``.
+        :type n: int
+        :param l: Angular momentum quantum number of the desired level. Default is ``0``.
+        :type l: int
+        :param verbose: Output diagnostic information. Default is ``False``.
+        :type verbose: boolean
+        :returns: Level population density of the desired level in units of cm^{-3}.
+        :rtype: float
+        """
+        assert type(ne) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert type(nHI) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert type(nHII) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert type(temp) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert (type(n) == int) and (n >= 3), "Invalid input for principle quantum number n."
+        assert (type(l) == int) and (l < n), "Invalid input for angular momentum quantum number l."
+        
+        try:
+            ne = ne.in_units('cm**(-3)').value
+            nHI = nHI.in_units('cm**(-3)').value
+            nHII = nHII.in_units('cm**(-3)').value
+            temp = temp.in_units('K').value
+        except TypeError:
+            pass
+        
         # check if all the quantities have the same dimension
         # convert LogT to float
-        LogT = float(LogT)
+        LogT = np.log10(temp)
         
         nmax     = self.nmax
         A        = self.A
         C        = self.C
         Alpha_nl = self.Alpha_nl
         q_nl     = self.q_nl
-        Config   = self.Config
+        Config   = self.config
 
         # test for consistency
         if (n < 1) or (n > self.nmax):
@@ -434,30 +472,55 @@ class HILevelPopulations:
                 conf_k = Config(n=nd, l=ld)
                 rhs    += A[conf_i][conf_k]
 
-        N       = np.zeros_like(LogT)
+        N       = np.zeros_like(LogT) * unyt.unyt_array(1, 'cm**(-3)')
         mask    = rhs > 0
         N[mask] = lhs[mask]/rhs[mask]
         if verbose:
             print("Computed level pop for level = {0:s}, log N = {1:2.4f}".format(conf_i, np.log10(N)))
         return N
 
-    def ComputeAllLevelPops(self, nHII = 1.0, ne = 1.0, nHI = 1.0, LogT=4.0):
-        '''
-        Compute level population for all levels  - implementents Eq. 4.10
-        Input: 
-           nHII(nH in the previous version) = proton number density nH [cm^[-3]]
-           nHI  = neutral hydrogen number density [cm^{-3}]
-           ne   = electron number density [[cm^-3]]
-           logT = logarithm of temperature
-        '''
+    def compute_all_level_pops(self, nHII = [1e2] * unyt.unyt_array(1, 'cm**(-3)'), 
+                                     ne = [1e2] * unyt.unyt_array(1, 'cm**(-3)'), 
+                                     nHI = [1e-5] * unyt.unyt_array(1, 'cm**(-3)'), 
+                                     temp = [1e4] * unyt.unyt_array(1, 'K'), verbose=False):
+        """
+        Compute level population for all levels.
+        
+        :param nHII: Proton number density [cm^{-3}]. Default is [1e2] * unyt.unyt_array(1, 'cm**(-3)').
+        :type nHII: float
+        :param ne: Electron number density [cm^{-3}]. Default is [1e2] * unyt.unyt_array(1, 'cm**(-3)').
+        :type ne: float
+        :param nHI: Neutral hydrogen number density [cm^{-3}]. Default is [1e-5] * unyt.unyt_array(1, 'cm**(-3)'). If collsional excitation is not enabled, any number input here will not go into the calculation. 
+        :type nHI: float
+        :param temp: Temperature. Default is [1e4] * unyt.unyt_array(1, 'K').
+        :type temp: float
+        :returns: Level population for all levels in units of cm^{-3}. 
+        :rtype: float
+        """
+        assert type(ne) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert type(nHI) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert type(nHII) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert type(temp) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        
+        
+        try:
+            ne = ne.in_units('cm**(-3)').value
+            nHI = nHI.in_units('cm**(-3)').value
+            nHII = nHII.in_units('cm**(-3)').value
+            temp = temp.in_units('K').value
+        except TypeError:
+            pass
+        
+        # check if all the quantities have the same dimension
+        # convert LogT to float
+        LogT = np.log10(temp)
         
-        #
         nmax     = self.nmax
         A        = self.A
         C        = self.C
         Alpha_nl = self.Alpha_nl
         q_nl     = self.q_nl
-        Config   = self.Config
+        Config   = self.config
         
 
         #
@@ -497,23 +560,26 @@ class HILevelPopulations:
 
                 N[conf_i] = 0.0
                 if rhs>0:
-                    N[conf_i] = lhs/rhs
+                    N[conf_i] = lhs/rhs * unyt.unyt_array(1, 'cm**(-3)')
         return N
         
-    def ComputeCascadeMatrix(self):
-        '''
-           Compute cascade matrix from Einstein coefficients
-        '''
-        import time
+    def compute_cascade_matrix(self):
+        """
+        Compute cascade matrix from Einstein coefficients
+        
+        :returns: Cascade matrix.
+        :rtype: dict
+        """
+        
         nmax     = self.nmax          # max upper level
         A        = self.A             # Einstein coefficient
         verbose  = self.verbose
-        Config   = self.Config
+        Config   = self.config
         topickle = True
         
         # if pickle file exists, read it
         if topickle:
-            import pickle
+        
             if self.caseB:
                 pname   = 'CascadeC_' + str(self.nmax) + '_' + 'B.pickle'
             else:
@@ -574,7 +640,7 @@ class HILevelPopulations:
                                 conf_k = Config(n=nd, l=ld)
                                 P[conf_i][conf_k] = A[conf_i][conf_k] / denom
         if self.verbose:
-            print(" ... Cascade matrix: probability matrix computed (eq. 4.8) in time {0:1.2f}".format(time.time()-t0))
+            print(" ... Cascade matrix: probability matrix computed in time {0:1.2f}".format(time.time()-t0))
         self.P = P
         
         # Compute the transpose of P
@@ -670,11 +736,15 @@ class HILevelPopulations:
         return C
 
                     
-    def ReadTabulatedEinsteinCoefficients(self, fname):
-        '''
-          Read tabulated Einstein coefficients
-          Use these to compute the casecade
-        '''
+    def read_tabulated_Einstein_coefficients(self, fname):
+        """
+        Read tabulated Einstein coefficients.
+        
+        :param fname: Filename.
+        :type fname: str
+        :returns: Dictionary of Einstein A values. 
+        :rtype: dict
+        """
         
         verbose = False  # set True to get timing info
         
@@ -696,9 +766,9 @@ class HILevelPopulations:
                     if self.verbose:
                         print(" ... Imposing caseB (no Lyman-transitions) ")
                         
-                    conf_k = self.Config(n=1, l=0) # ground state
+                    conf_k = self.config(n=1, l=0) # ground state
                     for nu in np.arange(2, self.nmax+1):
-                        conf_i = self.Config(n=nu, l=1) # all p-state
+                        conf_i = self.config(n=nu, l=1) # all p-state
                         A[conf_i][conf_k] = 0.0
                 return A
             else:
@@ -727,12 +797,12 @@ class HILevelPopulations:
         # loop over upper level
         for nu in np.arange(2, nmax+1):
             for lu in np.arange(nu):
-                conf_i = self.Config(n=nu, l=lu)
+                conf_i = self.config(n=nu, l=lu)
                 A[conf_i] = {}
                 # loop over lower level
                 for nd in np.arange(nu):
                     for ld in np.arange(nd):
-                        conf_k = self.Config(n=nd, l=ld)
+                        conf_k = self.config(n=nd, l=ld)
                         A[conf_i][conf_k] = 0
         t0 = time.time() - t0
         if verbose:
@@ -746,8 +816,8 @@ class HILevelPopulations:
         lds      = data['ld'][:]
         Avals    = data['A'][:]
         for nup, lup, nd, ld, Aval in zip(nups, lups, nds, lds, Avals):
-            conf_i = self.Config(n=nup, l=lup)
-            conf_k = self.Config(n=nd, l=ld)
+            conf_i = self.config(n=nup, l=lup)
+            conf_k = self.config(n=nd, l=ld)
             if nup <= nmax:
                 A[conf_i][conf_k] = Aval
             else:
@@ -761,8 +831,8 @@ class HILevelPopulations:
         lu = 0
         nd = 1
         ld = 0
-        conf_i = self.Config(n=nu, l=lu)
-        conf_k = self.Config(n=nd, l=ld)
+        conf_i = self.config(n=nu, l=lu)
+        conf_k = self.config(n=nd, l=ld)
         A[conf_i][conf_k] = A_2s_1s
 
         # Save the file first, before imposing Case B, otherwise the As are saved with wrong case
@@ -777,53 +847,176 @@ class HILevelPopulations:
         if self.caseB:
             if self.verbose:
                 print(" ... Imposing caseB (no Lyman-transitions) ")
-            conf_k = self.Config(n=1, l=0) # ground state
+            conf_k = self.config(n=1, l=0) # ground state
             for nu in np.arange(2, nmax+1):
-                conf_i = self.Config(n=nu, l=1) # p-state
+                conf_i = self.config(n=nu, l=1) # p-state
                 A[conf_i][conf_k] = 0.0
         
         return A
 
         
-    def Config(self, n=1, l=1):
-        '''
-              configuration states are tuples of the form (n,l), where:
-          n = principle quantum number, n=1->nmax
-          l = angular momentum number, l=0->n-1
-        '''
+    def config(self, n=1, l=1):
+        """
+        Configuration states are tuples of the form (n,l), where        
+        n = principle quantum number
+        l = angular momentum quantum number
+        
+        :param n: Principle quantum number. Default is ``1``.
+        :type n: int
+        :param l: Angular momentum quantum number. Default is ``1``.
+        :type l: int
+        :returns: Configuration of a specific state in tuple. 
+        :rtype: tuple
+        """
         return (n,l)
         
-    def DeConfig(self, config='1s'):
-        '''
-            extract n and l value for a given configuration state
-        '''
+    def de_config(self, config=(1,0)):
+        """
+        Extract n and l value for a given configuration.
+        
+        :param config: Configuration of a state.
+        :type config: tuple
+        :returns: Principle quantum number and angular momentum quantum number.
+        :rtype: int, int
+        """
         return config[0], config[1]
 
     def alpha_effective(self, n=2, l=0, LogT=4.0):
+        """
+        Effective recombination rate coefficient to a speciic level nl. 
+        
+        :param n: Principle quantum number. Default is ``2``.
+        :type n: int
+        :param l: Angular momentum quantum number. Default is ``0``.
+        :type l: int
+        :param LogT: Temperature in log10. Default is ``4.0``.
+        :type LogT: float
+        :returns: Effective recombination coefficient. 
+        :rtype: float
+        """
+        assert n >= 2, "Onlt n >= 2 makes sense."
         alpha_nl = 0 # based on Eq. 30 in Pengelly 1964 
-        conf_k = self.Config(n=n, l=l) # loop over all l-states within this upper level
+        conf_k = self.config(n=n, l=l) # loop over all l-states within this upper level
         # alpha * recom coeff
         for n in np.arange(n, self.nmax+1):
             for l in np.arange(n):
-                conf_i = self.Config(n=n, l=l)
+                conf_i = self.config(n=n, l=l)
                 if self.recom:
                     alpha_nl += 10**self.Alpha_nl[conf_i](LogT) * self.C[conf_i][conf_k]
         return alpha_nl
 
+    def get_A_coeffs(self, nupper=3, nlower=2):
+        '''
+        Summing Einstein coefficients for all substates in the upper level.
+
+        :param nupper: Principle quantum number for the upper level. Default is ``3``.
+        :type nupper: int
+        :param nlower: Principle quantum number for the lower level. Default is ``2``.
+        :type nlower: int
+        
+        :return: Summed Einstein coefficient.
+        :rtype: dict
+        '''
+        assert (type(nupper) == int) and (nupper >= 2) and (nupper > nlower), "Invalid input for nup."
+        assert (type(nlower) == int) and (nlower >= 1), "Invalid input for ndown."
+        
+        As = {}
+        for lup in np.arange(nupper):
+            conf_up  = self.config(n=nupper, l=lup)
+            A        = 0.0
+            for ldown in np.arange(nlower):
+                conf_down = self.config(n=nlower, l=ldown)
+                try:
+                    A += self.A[conf_up][conf_down]
+                except:
+                    pass
+            As[conf_up] = A
+        return As
+
+    def get_emissivity(self, ne=[1e2] * unyt.unyt_array(1, 'cm**(-3)'), 
+                             nHI=[1e-5] * unyt.unyt_array(1, 'cm**(-3)'), 
+                             nHII=[1e2] * unyt.unyt_array(1, 'cm**(-3)'), 
+                             temp=[1e4] * unyt.unyt_array(1, 'K'), 
+                             nupper=3, nlower=2):
+        '''Compute line emissivity for given density and temperature. 
+
+        :param ne: Electron density. Default is [1e2] * unyt.unyt_array(1, 'cm**(-3)'). 
+        :type ne: float
+        :param nHI: Neutral hydrogen density. Default is [1e-5] * unyt.unyt_array(1, 'cm**(-3)'). 
+        :type nHI: float
+        :param nHII: Proton density. Default is [1e2] * unyt.unyt_array(1, 'cm**(-3)'). 
+        :type nHII: float
+        :param temp: Temperature. Default is [1e4] * unyt.unyt_array(1, 'K'). 
+        :type: float
+        :param nupper: Principle quantum number for the upper level. Default is ``3``.
+        :type nupper: int
+        :param nlower: Principle quantum number for the lower level. Default is ``2``.
+        :type nlower: int 
+        :return: Line emissisity in erg s**(-1) cm**(-3) 
+        :rtype: float
+        '''
+        assert type(ne) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert type(nHI) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert type(nHII) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert type(temp) == unyt.array.unyt_array, 'Please input quantities with units using unyt_array.'
+        assert (type(nupper) == int) and (nupper >= 2) and (nupper > nlower), "Invalid input for principle quantum number nupper."
+        assert (type(nlower) == int) and (nlower >= 1), "Invalid input for principle quantum number nlower."
+        
+        try:
+            ne = ne.in_units('cm**(-3)').value
+            nHI = nHI.in_units('cm**(-3)').value
+            nHII = nHII.in_units('cm**(-3)').value
+            temp = temp.in_units('K').value
+        except TypeError:
+            pass
+        
+        population = np.zeros((len(ne), nupper), dtype=float) * unyt.unyt_array(1, 'cm**(-3)')
+        for l in np.arange(nupper):
+            population[:, l] = self.compute_level_pop(nHII=nHII * unyt.array.unyt_array(1, 'cm**(-3)'), 
+                                                   ne=ne * unyt.array.unyt_array(1, 'cm**(-3)'), 
+                                                   nHI=nHI * unyt.array.unyt_array(1, 'cm**(-3)'), 
+                                                   temp=temp * unyt.array.unyt_array(1, 'K'),  
+                                                   n=nupper, l=int(l), verbose=False)
+            
+        As = self.get_A_coeffs(nupper=nupper, nlower=nlower)
+    
+        emis = {}
+        emis_tot = 0
+        for lup in np.arange(nupper):
+            # get population level
+            conf        = self.config(n=nupper, l=lup)
+            emis[conf] = {}
+            emis[conf] = As[conf] * unyt.s**(-1) * self.Eion * (1./nlower**2 - 1./nupper**2) * population[:, lup]
+            emis_tot += emis[conf]
+            
+        return emis_tot.in_units('erg * cm**(-3) * s**(-1)')
+
+
     def branching_ratio(self, nupper=3, nlower=2, LogT=4.0, caseB=True):
+        '''
+        Calculate the fraction of recombination that results in the emission of a specific line, taking into account only radiative processes. 
+        The definition can be found in Liu et al. 2025. 
+
+        :param nupper: Principle quantum number for the upper level. Default is ``3``.
+        :type nupper: int
+        :param nlower: Principle quantum number for the lower level. Default is ``2``.
+        :type nlower: int
+        :return: The fraction of recombination that results in the emission of a specific line. 
+        :rtype: float
+        '''
         if caseB == True:
-            alpha_tot = self.AlphaB(LogT=LogT)
+            alpha_tot = self.alpha_B(LogT=LogT)
         else:
-            alpha_tot = self.AlphaA(LogT=LogT)
+            alpha_tot = self.alpha_A(LogT=LogT)
         
         lterms = np.zeros(nupper)
         
         As = {}
         for lup in np.arange(nupper):
-            conf_up  = self.Config(n=nupper, l=lup)
+            conf_up  = self.config(n=nupper, l=lup)
             Atemp    = 0.0
             for ldown in np.arange(nlower):
-                conf_down = self.Config(n=nlower, l=ldown)
+                conf_down = self.config(n=nlower, l=ldown)
                 try:
                     Atemp += self.A[conf_up][conf_down]
                 except:
@@ -832,35 +1025,31 @@ class HILevelPopulations:
 
         for lup in np.arange(nupper):
             lhs_rr = 0
-            conf_k = self.Config(n=nupper, l=lup) # loop over all l-states within this upper level
+            conf_k = self.config(n=nupper, l=lup) # loop over all l-states within this upper level
             # alpha * recom coeff
             for n in np.arange(nupper, self.nmax+1):
                 for l in np.arange(n):
-                    conf_i = self.Config(n=n, l=l)
+                    conf_i = self.config(n=n, l=l)
                     if self.recom:
                         lhs_rr += 10**self.Alpha_nl[conf_i](LogT) * self.C[conf_i][conf_k]
             
             rhs    = 0
-            conf_i = self.Config(n=nupper, l=lup)
+            conf_i = self.config(n=nupper, l=lup)
             
             for nd in np.arange(1, nupper):
                 for ld in [lup-1, lup+1]:
                     if (ld >=0) & (ld < nd):
-                        conf_k = self.Config(n=nd, l=ld)
+                        conf_k = self.config(n=nd, l=ld)
                         rhs += self.A[conf_i][conf_k]
                 if (nd == 1) & (nupper == 2) & (lup == 0):
                     ld     = 0
-                    conf_k = self.Config(n=nd, l=ld)
+                    conf_k = self.config(n=nd, l=ld)
                     rhs    += self.A[conf_i][conf_k]
             
             lterms[lup] = lhs_rr / rhs
-            
-        print('lterms', lterms)
 
         alpha_eff = 0
         for i, level_config in enumerate(As.keys()):
             alpha_eff += lterms[i] * As[level_config]
-            print('alpha_eff', alpha_eff, level_config)
-            print('alpha_a', alpha_tot)
         R = alpha_eff / alpha_tot
-        return alpha_eff, R
+        return R
